const AppDisplay   = imports.ui.appDisplay;
const AppFavorites = imports.ui.appFavorites.getAppFavorites();
const AppSystem    = imports.gi.Shell.AppSystem.get_default();
const GMenu        = imports.gi.GMenu;
const Lang         = imports.lang;
const Mainloop     = imports.mainloop;
const Me           = imports.misc.extensionUtils.getCurrentExtension();
const ViewSelector = imports.ui.main.overview._viewSelector;

const Prefs = Me.imports.prefs;

const SEARCH_CATEGORY = 'b88f323f-bff3-4106-8c21-de6974f38ede';

let providers = [];
let undos     = [];

function init() {
    let pos = let (
        menus = Prefs.getMenus()
    ) let (
        pos = menus.reduce(function(pos, menus, i, a) {
            menus.forEach(function(menu, j, a) {
                pos[menu] = { i : i, j : j };
            });
            return pos;
        }, {})
    ) function(dir) {
        if (dir.get_menu_id() in pos) {
            return pos[dir.get_menu_id()];
        }
        return { i : menus.length, j : 0 };
    };

    let hidden = let (
        hidden = Prefs.getHiddenApps().concat(Prefs.getMovedApps()).map(function(app, i, a) {
            return AppSystem.lookup_app(app);
        }).filter(function(app, i, a) {
            return app != null;
        })
    ) function(app) {
        return hidden.indexOf(app) >= 0 || app.get_app_info().get_nodisplay();
    };

    function walk(dir) {
        if (!dir.get_is_nodisplay()) {
            for (let dirs = [], apps = [], iter = dir.iter();;) {
                switch (iter.next()) {
                case GMenu.TreeItemType.DIRECTORY:
                    dirs = dirs.concat(walk(iter.get_directory()));
                    break;
                case GMenu.TreeItemType.ENTRY:
                    let app = AppSystem.lookup_app_by_tree_entry(iter.get_entry());
                    if (!hidden(app)) {
                        apps.push(app);
                    }
                    break;
                case GMenu.TreeItemType.INVALID:
                    if (apps.length > 0) {
                        dirs.push({
                            pos  : pos(dir),
                            dir  : dir,
                            apps : apps
                        });
                    }
                    return dirs;
                }
            }
        }
        return [];
    }

    function title(dirs) {
        let title = dirs.map(function(dir, i, a) {
            return dir.get_name().replace(/\s*&/g, ',');
        }).join(', ');

        let c = title.lastIndexOf(',');
        if (c > 0 && c < title.length - 2) {
            return title.substring(0, c) + ' &' + title.substring(c + 1);
        }
        return title;
    };

    function register(title, apps) {
        let provider = new AppDisplay.AppSearchProvider();

        provider.title = title;
        provider.getSubsearchResultSet = function(previousResults, terms) {
            provider.getInitialResultSet(terms);
        };
        provider.getInitialResultSet = function(terms) {
            if (terms == SEARCH_CATEGORY) {
                provider.searchSystem.pushResults(provider, apps);
            } else {
                provider.searchSystem.pushResults(provider, []);
            }

        };

        ViewSelector.addSearchProvider(provider);

        ViewSelector._searchResults._providerMeta.filter(function(meta, i, a) {
            return meta.provider === provider;
        }).forEach(function(meta, i, a) {
            meta.resultDisplay._grid._rowLimit = 10;
            if (Prefs.getCompactApplications()) {
                meta.resultDisplay._grid.actor.add_style_class_name('no-spacing');
            }
        });

        return provider;
    }

    // walk => [ { pos { i, j }, dir, apps }, ... ]
    providers = walk(AppSystem.get_tree().get_root_directory()).sort(function(a, b) {
        if (a.pos.i < b.pos.i || a.pos.i == b.pos.i && a.pos.j < b.pos.j) {
            return -1;
        }
        if (a.pos.i > b.pos.i || a.pos.i == b.pos.i && a.pos.j > b.pos.j) {
            return  1;
        }
        return a.dir.get_name().localeCompare(b.dir.get_name());
    }).reduce(function(groups, x, i, a) {
        if (typeof groups[x.pos.i] == 'undefined') {
            groups[x.pos.i] = {
                dirs : [],
                apps : []
            };
        }
        groups[x.pos.i].dirs.push(x.dir);
        groups[x.pos.i].apps = groups[x.pos.i].apps.concat(x.apps);
        return groups;
    }, []).map(function(group, i, a) {
        return {
            enable : function() {
                function appsSort(a, b) {
                    let aFav = AppFavorites.isFavorite(a.get_id());
                    let bFav = AppFavorites.isFavorite(b.get_id());

                    return aFav != bFav ? (aFav ? -1 : 1) : a.compare_by_name(b);
                }

                let provider = register(title(group.dirs), group.apps.sort(appsSort));

                function update() {
                    group.apps.sort(appsSort);

                    if (provider.searchSystem.getTerms() == SEARCH_CATEGORY) {
                        Mainloop.timeout_add(Prefs.getWorkaroundDelay(), function() {
                            provider.searchSystem.updateSearch(SEARCH_CATEGORY);
                        });
                    }
                }

                undos.push(
                    Prefs.connect(AppFavorites, 'changed', update),
                    Lang.bind(ViewSelector, ViewSelector.removeSearchProvider, provider)
                );
            }
        };
    });
}

function enable() {
    function state() {
        if (ViewSelector._searchSystem.getTerms() == SEARCH_CATEGORY) {
            return function() {
                ViewSelector._searchSystem.updateSearch(SEARCH_CATEGORY);
            };
        }
        return function() {};
    }

    function reset() {
        let restore = state();

        disable();
        init();
        enable();

        restore();
    }

    providers.forEach(function(provider, i, a) {
        provider.enable();
    });

    undos.push(
        Lang.bind(ViewSelector._searchResults, ViewSelector._searchResults.reset),
        Prefs.onAppsChange(reset),
        Prefs.onAppsStyleChange(reset),
        Prefs.connect(AppSystem, 'installed-changed', reset)
    );
}

function disable() {
    while(undos.length > 0) {
        undos.pop()();
    }
}
