const AppSystem = imports.gi.Shell.AppSystem.get_default();
const Lang      = imports.lang;
const Me        = imports.misc.extensionUtils.getCurrentExtension();
const Overview  = imports.ui.main.overview;
const PopupMenu = imports.ui.popupMenu;
const UserMenu  = imports.ui.main.panel.statusArea.userMenu.menu;

const Prefs = Me.imports.prefs;

let items = [];
let undos = [];

function init() {
    let position = UserMenu._getMenuItems().map(function(item, i, a) {
        return item instanceof PopupMenu.PopupSeparatorMenuItem;
    }).indexOf(true) + 1;

    items = Prefs.getMovedApps().map(function(app, i, a) {
        return AppSystem.lookup_app(app);
    }).filter(function(app, i, a) {
        return app != null;
    }).map(function(app, i, a) {
        return {
            enable : function() {
                let item = new PopupMenu.PopupMenuItem(app.get_name());
                UserMenu.addMenuItem(item, i + position);

                function activate() {
                    Overview.hide();
                    app.activate();
                }

                undos.push(
                    Prefs.connect(item, 'activate', activate),
                    Lang.bind(item, item.destroy)
                );
            }
        };
    });
}

function enable() {
    function reset() {
        disable();
        init();
        enable();
    }

    items.forEach(function(item, i, a) {
        item.enable();
    });

    undos.push(
        Prefs.onUserMenuChange(reset),
        Prefs.connect(AppSystem, 'installed-changed', reset)
    );
}

function disable() {
    while(undos.length > 0) {
        undos.pop()();
    }
}
